unit MainForm;
  (* This application allows the user to select a database and then to compile
  and display two tables (Tables and Fields) that together comprise a census of
  the tables and fields of the database.  The information displayed is a proper
  subset of what can be displayed using Borland's Database Explorer, but the
  utility of TableTable comes from the fact that each record of Tables and
  Fields includes a field in which the user can record comments to explain the
  structure of his or her database.

    The program has been tested only with databases of Paradox files, but it
  should work with others.  The tables it creates are Paradox files.

    The Paradox alias 'TableTable' is used to identify the directory in which
  the program's own DisplayHints table should be stored.

    When a database has been censused, the two tables are displayed in grids on
  the main form.  IndexTab controls facilitate browsing through the tables.
  The panels of the form are divided by splitters for user convenience.

    In the upper panel, the control to the right is a memo area in which to
  provide notes on the currently selected table.  The Sync button causes the
  current record of the fields table to be the first field of the currently
  selected table.

    In the lower panel, clicking on one of the colored column headers causes
  the fields table to be reindexed appropriately.  *)

  interface
    uses
      Buttons  (* tBitBtn *),
      Classes  (* tStringList *),
      ComCtrls,
      Controls  (* tCursor *),
      DBCtrls   (* tDBMemo, tDBNavigator, tDBText *),
      DBGrids   (* tDBGrid *),
      ExtCtrls  (* tPanel, tRadioGroup, tSplitter *),
      Forms     (* tForm, Screen *),
      Grids,
      IndexTab  (* tRRFIndexTab *),
      MyUtils   (* ModalShow, tRRFCloseButton *),
      RRFGrid   (* tRRFDBGrid *),
      RRFResize (* tRRFResizer *),
      StdCtrls  (* tComboBox, tLabel *),
      Windows;
      (* eject *)
    type
      tfmTableTable = class(tForm)
          btnCensus: tBitBtn;
          btnPrint:  tBitBtn;
          btnSync:   tBitBtn;
          cbxDataBase: tComboBox;
          gridTables: tDBGrid;
          dbmemoTableDescription: tDBMemo;
          navFields: tDBNavigator;
          navTables: tDBNavigator;
          dbtxTableName: tDBText;
          lbDataBase:    tLabel;
          lbFields:      tLabel;
          lbFieldsTable: tLabel;
          pnlFields: tPanel;
          pnlGrids:  tPanel;
          pnlHeader: tPanel;
          pnlTables: tPanel;
          rgDBType: tRadioGroup;
          tRRFCloseBtn1: tRRFCloseBtn;
          gridFields: tRRFDBGrid;
          idxFields:      tRRFIndexTab;
          idxFieldsTable: tRRFIndexTab;
          idxTablesTable: tRRFIndexTab;
          Resizer: tRRFResizer;
          splHeader_Grids:  tSplitter;
          splTables_Fields: tSplitter;
          procedure btnCensusClick   (Sender: tObject);
          procedure btnPrintClick    (Sender: tObject);
          procedure btnSyncClick     (Sender: tObject);
          procedure cbxDataBaseSelect(Sender: tObject);
          procedure FormShow         (Sender: tObject);
        private
          ChosenDataBaseName: string;
        end (* tfmTableTable *);

    var
      fmTableTable: tfmTableTable;

  implementation
    uses
      DB           (* tField, tFieldType, tIndexOptions *),
      DBTables     (* Session, tTable, tTableType *),
      MyConsts     (* nullstring *),
      Report       (* tfmReport *),
      TableTableDM (* dmTableTable *),
      Variants     (* VarArrayOf *);
      {$R *.dfm}

    const
      DescName = 'Description';

    resourcestring
      CaptionDoCensus   = 'Perform Census';
      CaptionUpCensus   = 'Update Census';
      NameTableOfFields = 'Fields';
      NameTableOfTables = 'Tables';
      (* eject *)
    procedure tfmTableTable.btnCensusClick(Sender: tObject);
      (* The user might want to update the census or to perform it
      anew.  Here, we create the files (if they don't already exist)
      and fill them, taking care not to disturb old comments *)

      const
        Required = true;

      var
        OldCursor: tCursor;
        TheTable: tTable;

      procedure AddAField(const FieldName: string;
                          IsRequired: boolean;
                          ADataType: tFieldType;
                          ASize: integer = 0);
        (* Create a field of the specified kind *)

        const
          StringSize = 60;

        begin (* AddAField *)
          with TheTable.FieldDefs.AddFieldDef do
            begin (* specify the field *)
              Name := FieldName;
              DataType := ADataType;
              Required := IsRequired;
              Size := ASize
            end   (* specify the field *)
        end   (* AddAField *);
        (* eject *)
      procedure CreateFieldsTable;
        (* Create the Fields table, which for each field of each table will
        have a record showing
            table name
            field name
            field type
            required?
            size
            description
        *)

        procedure FieldsIndex(const IndexName, FieldList: string);
          (* Create a new index with the specified properties *)

          begin (* FieldsIndex *)
            with TheTable.IndexDefs.AddIndexDef do
              begin (* set properties *)
                Name := IndexName;
                Fields := FieldList;
                if IndexName = nullstring then
                  Options := [ixPrimary, ixUnique]
                else
                  Options := [ixUnique]
              end   (* set properties *)
          end   (* FieldsIndex *);
          
        begin (* CreateFieldsTable *)
          TheTable := dmTableTable.tbFields;
          with TheTable do
            if not Exists then
              begin (* try to create it *)
                Close;
                FieldDefs.Clear;
                AddAField('TableName', Required,     ftString, 40);
                AddAField('FieldName', Required,     ftString, 40);
                AddAField('FieldType', Required,     ftString, 30);
                AddAField( 'Required', Required,     ftBoolean);
                AddAField(   'Size'  , not Required, ftSmallInt);
                AddAField(  DescName , not Required, ftString, 255);

                IndexDefs.Clear;
                FieldsIndex(    nullstring    ,        'TableName;FieldName'        );
                FieldsIndex(   'byFieldName'  ,        'FieldName;TableName'        );
                FieldsIndex(   'byFieldType'  ,    'FieldType;FieldName;TableName'  );
                FieldsIndex('bySizedFieldType', 'FieldType;Size;FieldName;TableName');

                CreateTable
              end   (* try to create it *)
        end   (* CreateFieldsTable *);
        (* eject *)
      procedure CreateTablesTable;
        (* Create the Tables table, which for each table will have a record
        showing
            table name
            record count
            description
        *)

        begin (* CreateTablesTable *)
          TheTable := dmTableTable.tbTables;
          with TheTable do
            if not Exists then
              begin (* try to create it *)
                Close;
                DatabaseName := ChosenDataBaseName;
                FieldDefs.Clear;
                AddAField( 'TableName' , Required,     ftString, 40);
                AddAField('RecordCount', Required,     ftInteger);
                AddAField(   DescName  , not Required, ftMemo, 240);

                IndexDefs.Clear;
                with IndexDefs.AddIndexDef do
                  begin (* primary index by table *)
                    Name := nullstring;
                    Fields := 'TableName';
                    Options := [ixPrimary, ixUnique]
                  end   (* primary index by table *);

                CreateTable;
                Open;

                with gridTables do
                  begin (* attach columns *)
                    Columns[0].FieldName := 'TableName';
                    Columns[1].FieldName := 'RecordCount'
                  end   (* attach columns *);
                dbmemoTableDescription.DataField := 'Description'
              end   (* try to create it *)
        end   (* CreateTablesTable *);
        (* eject *)
      procedure FillTables;
        (* for every table TTT in the database
             make (or update) an entry in the Tables table
             for every field in TTT
               make (or update) an entry in the Fields table
        *)

        const
          dbtASCII   = 0 (* from radio group *);
          dbtDBase   = 1;
          dbtDefault = 2;
          dbtFoxPro  = 3;
          dbtParadox = 4;
          dbtSQL     = 5;

        type
          dbtRange = dbtASCII .. dbtSQL;
          dbtBool    = array[dbtRange] of boolean;
          dbtStrings = array[dbtRange] of string;
          dbtType    = array[dbtRange] of tTableType;

          (* ASCII    DBase     default     FoxPro    Paradox    SQL  *)
        const
          FileFilter: dbtStrings =
            ('*.TXT', '*.DBF', nullstring, nullstring, '*.DB', nullstring);
          UseExtensions: dbtBool =
            (false,    true,     false,      false,     true,    false);
          PullSysTables: dbtBool =
            ( false,   false,    false,      false,     false,   true);
          TheType: dbtType =
            (ttASCII, ttDBase, ttDefault,  ttFoxPro,  ttParadox, ttDefault);

        var
          I: integer;
          TableNames: tStringList;
          Table: tTable;
          (* eject *)
        procedure ProcessTable(const ATableName: string);
          (* make (or update) an entry in the Tables table
              table name
              record count
             for every field in TTT
               make (or update) an entry in the Fields table
          *)

          var
            I: integer;

          procedure ProcessField(TheField: tField);
            (* make (or update) an entry in the Fields table *)

            const
              FieldTypeName: array[tFieldType] of string =
                ('unknown',  'string',      'smallint', 'integer',   'word',
                 'boolean',  'float',       'currency', 'BCD',       'date',
                 'time',     'DateTime',    'bytes',    'VarBytes',  'AutoInc',
                 'blob',     'memo',        'graphic',  'FmtMemo',   'ParadoxOle',
                 'DBaseOle', 'TypedBinary', 'cursor',   'FixedChar', 'WideString',
                 'largeint', 'ADT',         'array',    'reference', 'dataset',
                 'OraBlob',  'OraClob',     'variant',  'interface', 'IDispatch',
                 'GUID',     'timestamp',   'FMTBcd');

            begin (* ProcessField *)
              with dmTableTable.tbFields do
                begin (* with *)
                  if not Locate('TableName;FieldName',
                                VarArrayOf([Table.TableName, TheField.FieldName]),
                                []) then
                    begin (* new record *)
                      Insert;
                      FieldByName('TableName').AsString := Table.TableName;
                      FieldByName('FieldName').AsString := TheField.FieldName
                    end   (* new record *)
                  else
                    Edit;

                  with FieldByName('FieldType') do
                    AsString := FieldTypeName[TheField.DataType];
                  FieldByName('Required').AsBoolean := TheField.Required;
                  FieldByName(  'Size'  ).AsInteger := TheField.Size;

                  Post
                end   (* with *)
            end   (* ProcessField *);
            (* eject *)
          begin (* ProcessTable *)
            with dmTableTable do
              begin (* data module *)
                with Table do
                  begin (* scratch table *)
                    Close;
                    TableName := ATableName;
                    Open
                  end   (* scratch table *);

                with tbTables do
                  begin (* table of tables *)
                    if not Locate('TableName', ATableName, []) then
                      begin (* new record *)
                        Insert;
                        FieldByName('TableName').AsString := ATableName
                      end   (* new record *)
                    else
                      Edit;

                    FieldByName('RecordCount').AsInteger := Table.RecordCount;
                    Post
                  end   (* table of tables *);

                with Table do
                  for I := 1 to FieldCount do
                    ProcessField(Fields[I-1])
              end   (* data module *)
          end   (* ProcessTable *);
          (* eject *)
        begin (* FillTables *)
          dmTableTable.InsertionOK := true;
          Session.Active := true;
          TableNames := tStringList.Create;

          try     (* list must be freed *)
            Table := tTable.Create(nil);

            try (* table must be freed *)
              Table.TableType := TheType[rgDBType.ItemIndex];
              Table.DatabaseName := ChosenDataBaseName;

              with dmTableTable.tbTables.DBSession do
                GetTableNames(ChosenDataBaseName,
                              FileFilter   [rgDBType.ItemIndex],
                              UseExtensions[rgDBType.ItemIndex],
                              PullSysTables[rgDBType.ItemIndex],
                              TableNames);

              for I := 1 to TableNames.Count do
                ProcessTable(TableNames.Strings[I-1])
            finally (* table must be freed *)
              Table.Free
            end     (* table must be freed *)
          finally (* list must be freed *)
            TableNames.Free;
            Session.Active := false;
            dmTableTable.InsertionOK := false
          end     (* list must be freed *);

          pnlGrids.Visible := true
        end   (* FillTables *);

      begin (* tfmTableTable.btnCensusClick *)
        OldCursor := Screen.Cursor;
        Screen.Cursor := crHourGlass;

        with dmTableTable do
          begin (* data module *)
            tbTables.DisableControls;
            tbFields.DisableControls;
            try     (* must restore cursor *)
              if not tbTables.Exists then
                CreateTablesTable;
              tbTables.Open;

              if not tbFields.Exists then
                CreateFieldsTable;
              tbFields.Open;

              FillTables;
              tbTables.Open;
              tbFields.Open;
              tbTableFields.Open
            finally (* must restore cursor *)
              tbFields.EnableControls;
              tbTables.EnableControls;
              Screen.Cursor := OldCursor
            end     (* must restore cursor *)
          end   (* data module *)
      end   (* tfmTableTable.btnCensusClick *);
      (* eject *)
    procedure tfmTableTable.btnPrintClick(Sender: tObject);
      (* Print report *)

      begin (* tfmTableTable.btnPrintClick *)
        ModalShow(tfmReport)
      end   (* tfmTableTable.btnPrintClick *);

    procedure tfmTableTable.btnSyncClick(Sender: tObject);
      (* Makes Fields grid focus on Table selected in Tables grid *)

      begin (* tfmTableTable.btnSyncClick *)
        with dmTableTable do
          with tbTables do
            if RecordCount > 0 then
              if not tbFields.Locate('TableName',
                                     FieldByName('TableName').AsString,
                                     []) then
                (* ignore it *)
      end   (* tfmTableTable.btnSyncClick *);

    procedure tfmTableTable.cbxDataBaseSelect(Sender: tObject);
      (* Database selected *)

      procedure OfferToCreate;
        (* The table of tables doesn't exist.  Make the button be a
        creator *)

        begin (* OfferToCreate *)
          pnlGrids.Visible := false;
          with btnCensus do
            begin (* button *)
              Caption := CaptionDoCensus;
              Visible := true
            end   (* button *)
        end   (* OfferToCreate *);

      procedure UseTables;
        (* The table of tables exists.  Open it, but use the button to
        refresh *)

        begin (* UseTables *)
          with btnCensus do
            begin (* button *)
              Caption := CaptionUpCensus;
              Visible := true
            end   (* button *);

          pnlGrids.Visible := true;
          with dmTableTable do
            begin (* use data module *)
              tbTables.Open;
              tbFields.Open
            end   (* use data module *)
        end   (* UseTables *);
        (* eject *)
      begin (* tfmTableTable.cbxDataBaseSelect *)
        with cbxDataBase do
          ChosenDataBaseName := Items[ItemIndex];

        with dmTableTable do
          begin (* use data module *)
            with tbTables do
              begin (* table of tables *)
                Close;
                DataBaseName := ChosenDataBaseName;
                TableName := NameTableOfTables
              end   (* table of tables *);

            with tbFields do
              begin (* table of fields *)
                Close;
                DataBaseName := ChosenDataBaseName;
                TableName := NameTableOfFields
              end   (* table of fields *);

            with tbTableFields do
              begin (* slave table for report printer *)
                Close;
                DataBaseName := ChosenDataBaseName;
                TableName := NameTableOfFields
              end   (* slave table for report printer *);

            if tbTables.Exists and tbFields.Exists then
              UseTables
            else
              OfferToCreate
          end   (* use data module *)
      end   (* tfmTableTable.cbxDataBaseSelect *);
      (* eject *)
    procedure tfmTableTable.FormShow(Sender: tObject);
      (* Initialization. The ShowResizerButton constant here could
      more elegantly be replaced by an INI-dependent switch that
      allowed the user to hide the button once display hints had been
      recorded for each of the user's preffered display resolutions. *)

      const
        ShowResizerButton = true;

      procedure ListDataBases;
        (* Fill the combo box with database names *)

        var
          I: integer;
          List: tStringList;

        begin (* ListDataBases *)
          Session.Active := true;
          List := tStringList.Create;

          try     (* list must be freed *)
            Session.GetDatabaseNames(List);
            with cbxDataBase do
              for I := 1 to List.Count do
                AddItem(List.Strings[I-1], nil)
          finally (* list must be freed *)
            List.Free;
            Session.Active := false
          end     (* list must be freed *)
        end   (* ListDataBases *);

      begin (* tfmTableTable.FormShow *)
        Resizer.UseDisplayHints(self, ShowResizerButton);
        btnCensus.Visible := false;
        pnlFields.Caption := nullstring;
        pnlGrids.Caption := nullstring;
        pnlGrids.Visible := false;
        pnlHeader.Caption := nullstring;
        pnlTables.Caption := nullstring;
        idxFields.Arm;
        idxFieldsTable.Arm;
        idxTablesTable.Arm;
        ListDataBases
      end   (* tfmTableTable.FormShow *);

end.
